/* ------------------------------------------------------------------
 * Un exemple d'implmentation de malloc()/free()
 * Jean-Paul Rigault
 * ------------------------------------------------------------------
 * Implmentation
 * ------------------------------------------------------------------
 */

#include <assert.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#include "mymalloc.h"
#include "mymalloc_impl.h"

/* Liste des blocks libres
 * -----------------------
 *
 * Les blocs libres sont chans entre eux dans une liste circulaire simplement
 * chaine, range dans l'ordre des adresses machines croissantes. Les
 * chanages sont effectus par l'intermdiaire de l'entte de gestion des
 * blocs. Le dbut (et la fin, puisqu'elle est circulaire !) de cette liste est
 * indiqu par une "sentinelle", c'est--dire un bloc "bidon" qui ne correspond
 * pas  un vritable bloc de mmoire alloue. Cette sentinelle est toujours
 * prsente, et est nomme ici base. Par commodit, le pointeur pfree contient
 * l'adresse de cette base (il reste constant).
 *
 * Au dpart la liste de blocs libres est vide, donc base pointe sur elle-mme.
 */
static struct Header base = {0, &base};	/* sentinelle de la liste circulaire de
                                         * blocs libres */
static struct Header *pfree = &base;	/* pointeur (constant) sur cette sentinelle */

/* Fonction d'allocation de mmoire systme (definie plus loin)
* -------------------------------------------------------------
*
* Le paramtre est le nombre d'enttes
*/
static void *system_alloc(unsigned int nh);

/* Fonction d'allocation mymalloc(sz)
 * ----------------------------------
 *
 * Elle reoit la taille sz (en caractres)  allouer. Elle calcule le nombre
 * d'enttes (nh) ncessaires pour allouer cette taille (plus une entte
 * servant  la gestion). Puis elle parcourt la liste libre pour trouver le
 * premier block de taille suffisante. Si elle trouve, elle dcoupe ce block
 * pour ne retenir que ce dont elle a besoin et recycle le reste dans la liste
 * libre. Si elle ne trouve pas elle appelle system_alloc() pour demander de
 * la mmoire supplmentaire au systme, place la mmoire obtenue dans la liste
 * libre et continue sa recherche (qui sera alors immdiatement fructueuse).
 *
 * Le pointeur retourn pointe sur la premire adresse "utile" (c'est--dire
 * utilisable pour les donnes de l'appelant) du block ainsi allou.
 */
void *mymalloc(unsigned int sz)
{
    unsigned int nh; 		/* taille ncessaire en nombre d'enttes */
    struct Header *ph;		/* Pointeur sur le bloc libre courant */
    struct Header *phprev;	/* Pointeur sur le bloc libre prcdent */

    /* On doit allouer suffisamment d'enttes pour contenir la taille sz plus 1
     * pour assurer la gestion
     */
    nh = NBLOCKS(sz, HEADER_SZ) + 1;

    /* Parcours de la liste libre : on doit garder un pointeur sur le bloc
     * prcdent
     */
    for (phprev = pfree, ph = pfree->next;  ; phprev = ph, ph = ph->next)
    {
        if (ph->sizenh >= nh) 
        {
            /* Le block courant est de taille suffisante */
            if (ph->sizenh == nh)
            {
                /* Ici, c'est mme *exactement* la taille ncessaire, on retire
                 * ce block de la liste libre en le "shuntant"
                 */
                phprev->next = ph->next;

                --nfree;	/* un bloc libre de moins */
            }
            else
            {
                /* Ici le block trouv est trop grand, on le dcoupe, la partie
                 * utile tant  la fin et le dbut tant recycl dans la liste
                 * libre
                 */
                ph->sizenh -= nh;	/* on prend ce dont on a besoin */
                ph += ph->sizenh; 	/* on saute le dbut qui va rester libre */
                ph->sizenh = nh;  	/* ph pointe maintenant sur l'entte de
                                         * gestion du bloc que l'on va retourner,
                                         * on ajuste nh pour un futur myfree
					 * (la valeur de next est indiffrente) */

                /* Ici on n'a pas  mettre  jour nfree, car il reste un bloc
                 * libre
                 */
            }

            /* Instrumentation */
            ++nuserblk;		/* un bloc utilisateur de plus */
            totalnh += nh;	/* blocs d'entte supplmentaires */
            usersz += sz;	/* taille utile pour l'utilisateur */

            /* Dans tous les cas de succs, on retourne un pointeur sur la
             * premire adresse utile (c'est--dire juste aprs l'entte de
             * gestion) */
            return ph + 1;
        }

        /* Le block libre courant n'est pas assez grand, ou alors il n'y  a
         * plus de bloc libre.
         */

        if (ph == pfree)
        {
            /* On a fait un tour complet de la liste libre sans rien
             * trouver. On demande donc assez de mmoire au systme pour
             * allouer nos sizenh
             */
            if ((ph = system_alloc(nh)) == NULL) 
            {
                /* On ne peut plus rien obtenir */
                fprintf(stderr, "mymalloc: error: no more system memory\n");
                return NULL;
            }
            /* On a pu allouer de la mmoire systme et on peut donc continuer
             * notre parcours de la liste libre qui va maintenant, par
             * construction, trouver un bloc de taille suffisante.
             * 
             * Noter que system_alloc() retourne l'adresse du bloc  partir
             * duquel le parcours peut continuer.
             */
        }

        /* Continuation de la boucle de parcours des blocks libres */
    }
    /*NOTREACHED*/
}

/* Fonction pour rendre de la mmoire myfree(ptr)
 * ----------------------------------------------
 *
 * Le paramtre ptr est un pointeur sur la premire adresse utile du bloc
 * (i.e., celle qui suit l'entte de gestion). Bien entendu p doit avoir une
 * valeur retourne par un prcdent appel de mymalloc(). Librer plusieurs
 * fois le mme bloc peut tre catastrophique.
 *
 * On parcourt la liste de blocs libres pour trouver le premier (rappelons-nous
 * que la liste des blocs libres est trie par adresses croissantes) dont
 * l'adresse est suprieure  celle du bloc que l'on libre. Si les deux blocs
 * sont adjacents (pour les adresses machines), on constitue un gros bloc libre
 * par fusion. De mme si le bloc  librer est immdiatement prcd (toujours
 * pour les adresses machines) d'un bloc libre, on fusionne galement. Dans
 * tous les cas le bloc ainsi obtenu est remis dans la liste libre.
 */
void myfree(void *ptr)
{
    struct Header *ptrh;	/* le dbut du bloc correspondant  l'adresse utile ptr */
    struct Header *ph;		/* Pointeur sur le bloc libre courant */
    struct Header *phprev;	/* Pointeur sur le bloc libre prcdent */

    /* ptr tant l'adresse utile, le dbut du bloc commence un entte avant */
    ptrh = (struct Header *)ptr - 1;
    
    /* Parcours de la liste libre afin de trouver le bloc juste suivant. */
    for (phprev = pfree, ph = pfree->next;  ; phprev = ph, ph = ph->next)
    {
        /* Soit on fait un tour complet sans rien trouver (p == pfree), soit le
         * bloc courant suit immdiatement le bloc  librer
         */
        if (ph == pfree || ph > ptrh) break;
    }

    /* Ici, deux cas sont possibles :
     *
     * Soit on a trouv un bloc libre dont l'adresse suit celle du bloc 
     * librer. Dans ce cas ph pointe sur ce bloc libre et phprev pointe sur le
     * dernier bloc libre dont l'adresse prcde celle du bloc  librer (ou
     * bien phprev == pfree, auquel cas cela signifie que le bloc  librer va
     * se placer en tte de la liste libre).
     *
     * Soit on n'a pas trouv de tel bloc et alors ph == pfree et phprev pointe
     * sur le dernier bloc libre de la liste (ou phprev == pfree si la liste en
     * question est vide). Dans ce cas le bloc  librer doit devenir le
     * dernier de la liste.
     *
     * On voit donc que *dans tous les cas* les deux pointeurs phprev et ph
     * "encadrent" la position o l'on doit placer le bloc  librer dans la
     * liste.
     */

    /* Instrumentation */
    ++nfree;			/* un bloc libre de plus */
    --nuserblk;			/* un bloc utilisateur de moins */
    totalnh -= ptrh->sizenh;	/* on rend ces enttes */
    
    /* On essaie de fusionner avec le bloc suivant */
    if (ptrh + ptrh->sizenh == ph) 
    {
        /* Le bloc  librer et le bloc libre qui suit sont adjacents */
        ptrh->sizenh += ph->sizenh;	/* somme des taille */
        ptrh->next = ph->next;		/* maintien de la liste libre */

        --nfree;	/* un bloc libre de moins */
    }
    else 
    {
        /* Pas de fusion avec le bloc libre qui suit */
        ptrh->next = ph;		/* maintien de la liste libre */
    }

    /* On essaie de fusionner avec le bloc prcdent */
    if (phprev + phprev->sizenh == ptrh)
    {
         /* Le bloc  librer et le bloc libre qui prcde sont adjacents */
        phprev->sizenh += ptrh->sizenh; 	/* somme des tailles */
        phprev->next = ptrh->next;		/* maintien de la liste libre */

        --nfree;	/* un bloc libre de moins */    }
    else 
    {
         /* Pas de fusion avec le bloc libre qui prcde */
        phprev->next = ptrh;		/* maintien de la liste libre */
    }
}

/* Fonction d'allocation de mmoire systme
* -----------------------------------------
*
* Le paramtre est le nombre d'enttes. La veleur de retour est l'adresse 
* partir de laquelle le parcours de la liste libre peut continuer dans
* mymalloc().
*/
static void *system_alloc(unsigned int nh)
{
    struct Header *phnew;	/* pointeur sur la nouvelle zone mmoire */

    /* Si la taille demande est infrieure au minimum, on demandera le
     * minimum
     */
    if (nh < MINSYSTBLOCK_NH) nh = MINSYSTBLOCK_NH;

    /* Demande de mmoire au systme (l'argument de sbrk() est en nombre de
     * caractres)
     */
    phnew = (struct Header *)sbrk(nh * HEADER_SZ);
    if ((int)phnew < 0) return NULL; 	/* plus de mmoire */

    /* Instrumentation */
    ++nsysblk;		/* un bloc systme de plus */
    syssz += nh * HEADER_SZ;

    /* Mise  jour des informations de gestion du nouveau bloc */
    phnew->sizenh = nh;

    /* On doit placer le nouveau bloc dans la liste de blocs libres, 
     * l'emplacement correspondant  son adresse machine. Le plus simple est de
     * faire comme s'il s'agissait d'un bloc qu'on "libre" et donc d'appeler
     * myfree() qui le placera  l'endroit correct de la liste libre.
     *
     * Mais attention ! La fonction myfree() doit tre invoque avec l'adresse
     * "utile" du bloc, c'est--dire l'adresse qui suit l'entte de gestion
     * (d'o le + 1 qui suit)
     */
    ++nuserblk;	/* c'est comme si il y avait un bloc utilisateur de plus */
    totalnh += nh;	/* et autant d'enttes supplmentaires */
    myfree(phnew + 1);

    /* Nous n'avons aucune garantie sur la position de l'adresse du nouveau
     * bloc (phnew) par rapport  celle des blocs dj existants. Il faut donc
     * que le parcours de la liste libre reparte du dbut.
     */
    return pfree;    
}

/* Fonction d'instrumentation
 * --------------------------
 *
 * sysblk	nombre de blocks systme obtenus pas sbrk() [ne dcroit pas]
 * syssz	taille totale (en caractres) obtenue par sbrk() [ne dcroit pas] 
 * userblk	nombre de blocks couramment allous par l'appelant (incrment
 *		de 1 par mymalloc(), dcrment de 1 par myfree())
 * totalsz	taille totale (en caractres) alloue pour l'appelant (c'est la
 *		taille totale des blocs prcdents
 * nfree	nombre de blocs actuellement dans la liste libre
 */
void mymalloc_instrum()
{
    fprintf(stderr, "*** sysblk = %d syssz = %d usersz = %d\n"
            "*** userblk = %d totalsz = %d nfree = %d\n",
            nsysblk, syssz, usersz, 
            nuserblk, totalnh * HEADER_SZ, nfree);
}
